# loadAndStoreNumpy.py loads file given on command lines and saves using numpy
# D. Parson Fall 2023 CSCx23

import sys
import numpy as np
import csv
from pprint import pformat

if __name__ == '__main__':  # run from command line
    dataset = np.genfromtxt(sys.argv[1], delimiter=",", filling_values=None,
        names=True, dtype=None)
    # https://numpy.org/doc/stable/user/how-to-io.html
    # https://numpy.org/doc/stable/reference/generated/numpy.genfromtxt.html
    # We can use -99999 instead of None because we do not want to force
    # floating point type on int columns by using np.nan or object
    # instead of a numeric tyep.
    print('dtype', (dataset.dtype))
    print('First row: ', pformat(list(dataset[0]),width=80))
    print('Last row: ', pformat(list(dataset[-1]),width=80))
    listOfTuples = eval(str(dataset.dtype))
    print('Formatted dtype', pformat(listOfTuples, width=80))
    print('shape',dataset.shape)
    header = [tup[0] for tup in listOfTuples]

    # Database select:
    yearcol = header.index('year')
    monthcol = header.index('month')
    def filtFunc(yearcol,monthcol):
        # This is a closure that binds yearcol,monthcol in the outer function.
        def f(row):
            try:
                return row[yearcol] == 1994 and row[monthcol] == 10
            except:
                return False
        return f
    print()
    for row in filter(filtFunc(yearcol,monthcol), dataset):
        print('SELECTION',row)
    print('')

    # Database project of columns [0], [1], [-1] (last)
    # https://numpy.org/doc/stable/reference/generated/numpy.take.html
    # col_dataset = np.asfortranarray(dataset)
    # projection = np.take(col_dataset, indices=[0, 1]) #, len(header)-1], axis=None)
    # ^^^ NIGHTMARE. It keeps grabbing rows. I want [0], [1], and last col
    projection = [(row[0], row[1], row[-1]) for row in dataset]
    # They are in random order. Sort by year, month
    # NO! PRESERVE ROW ORDER UNTIL JOIN. sorted() is not a mutator.
    # projection.sort(key=lambda row : (row[0], row[1]))
    sortedTriplet = sorted(projection, key=lambda row : (row[0], row[1]))
    print('PROJECTION of [0, 1, -1]:\n',pformat(projection), '\n')

    # Database join of two projections.
    HMtempC_mean_col = header.index('HMtempC_mean')
    WindSpd_mean_col = header.index('WindSpd_mean')
    wnd_WNW_NW_col = header.index('wnd_WNW_NW')
    # Find min and max per columns of interest for [0.0, 1.0] normalization.
    minmax = [[dataset[0][HMtempC_mean_col], dataset[0][HMtempC_mean_col]],
        [dataset[0][WindSpd_mean_col], dataset[0][WindSpd_mean_col]],
        [dataset[0][wnd_WNW_NW_col], dataset[0][wnd_WNW_NW_col]]]
    for row in dataset:
        minmax[0][0] = min(minmax[0][0], row[HMtempC_mean_col])
        minmax[0][1] = max(minmax[0][1], row[HMtempC_mean_col])
        minmax[1][0] = min(minmax[1][0], row[WindSpd_mean_col])
        minmax[1][1] = max(minmax[1][1], row[WindSpd_mean_col])
        minmax[2][0] = min(minmax[2][0], row[wnd_WNW_NW_col])
        minmax[2][1] = max(minmax[2][1], row[wnd_WNW_NW_col])
    joinedDataset = []
    for rowindex in range(0, len(dataset)):
        try:
            # scale each value the percentage of the way from its min to max
            norm_HMtempC_mean_col = round(((dataset[rowindex][HMtempC_mean_col]
                - minmax[0][0]) / (minmax[0][1] - minmax[0][0])),6)
        except:
            norm_HMtempC_mean_col = np.nan
        try:
            norm_WindSpd_mean_col = round(((dataset[rowindex][WindSpd_mean_col]
                - minmax[1][0]) / (minmax[1][1] - minmax[1][0])),6)
        except:
            norm_WindSpd_mean_col = np.nan
        try:
            norm_wnd_WNW_NW_col = round(((dataset[rowindex][wnd_WNW_NW_col]
                - minmax[2][0]) / (minmax[2][1] - minmax[2][0])),6)
        except:
            norm_wnd_WNW_NW_col = np.nan
        joinedDataset.append(projection[rowindex]
            + (norm_HMtempC_mean_col, norm_WindSpd_mean_col,
                norm_wnd_WNW_NW_col))
    jhdr = (header[0], header[1], header[-1],
        header[HMtempC_mean_col], header[WindSpd_mean_col],
        header[wnd_WNW_NW_col])
    print('JOIN NORMALIZED: ', jhdr)
    joinedDataset.sort(key=lambda row : (row[0], row[1]))
    print(pformat(joinedDataset), '\n')


    extensionIndex = sys.argv[1].find('.csv')
    outname = sys.argv[1][0:extensionIndex] + '_Numpy' + '.csv'
    # np.savetxt(outname, dataset, delimiter=',', newline='\n')
    # ^^^ THAT LINES LOSES HEADERS, SAVES THEM ALL AS FLOATS
    ofile = open(outname, 'w')
    ocsv = csv.writer(ofile)
    ocsv.writerow(header)
    ocsv.writerows(dataset)
    ofile.close()
